/*
    gen_html.c - Part of psiconv, a PSION 5 file formats converter
    Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "psiconv/data.h"
#include "psiconv/list.h"
#include "gen.h"
#include "psiconv.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif

/* This determines for each character how it is displayed */
static const char *char_table[0x100] =
{ 
  /* 0x00 */ ""     ,""     ,""     ,""     ,""     ,""     ,"<P>  ","<BR>"  ,
  /* 0x08 */ "<P>"  ," "    ,""     ,""     ,""     ,""     ,""     ,""      ,
  /* 0x10 */ " "    ,""     ,""     ,""     ,""     ,""     ,""     ,""      ,
  /* 0x18 */ ""     ,""     ,""     ,""     ,""     ,""     ,""     ,""      ,
  /* 0x20 */ " "    ,"!"    ,"&quot;","#"   ,"$"    ,"%"    ,"&amp;","'"     ,
  /* 0x28 */ "("    ,")"    ,"*"    ,"+"    ,","    ,"-"    ,"."    ,"/"     ,
  /* 0x30 */ "0"    ,"1"    ,"2"    ,"3"    ,"4"    ,"5"    ,"6"    ,"7"     ,
  /* 0x38 */ "8"    ,"9"    ,":"    ,";"    ,"&lt;"  ,"="   ,"&gt;" ,"?"     ,
  /* 0x40 */ "@"    ,"A"    ,"B"    ,"C"    ,"D"    ,"E"    ,"F"    ,"G"     ,
  /* 0x48 */ "H"    ,"I"    ,"J"    ,"K"    ,"L"    ,"M"    ,"N"    ,"O"     ,
  /* 0x50 */ "P"    ,"Q"    ,"R"    ,"S"    ,"T"    ,"U"    ,"V"    ,"W"     ,
  /* 0x58 */ "X"    ,"Y"    ,"Z"    ,"["    ,"\\"   ,"]"    ,"^"    ,"_"     ,
  /* 0x60 */ "`"    ,"a"    ,"b"    ,"c"    ,"d"    ,"e"    ,"f"    ,"g"     ,
  /* 0x68 */ "h"    ,"i"    ,"j"    ,"k"    ,"l"    ,"m"    ,"n"    ,"o"     ,
  /* 0x70 */ "p"    ,"q"    ,"r"    ,"s"    ,"t"    ,"u"    ,"v"    ,"w"     ,
  /* 0x78 */ "x"    ,"y"    ,"z"    ,"{"    ,"|"    ,"}"    ,"~"    ,""      ,
  /* 0x80 */ ""     ,"","&sbquot;","&fnof;","&bdquo;","&hellip;",
                                                        "&dagger;","&Dagger;",
  /* 0x88 */ "^","&permil;","&Scaron;","&lang;","&OElig;","" ,""    ,""      ,
  /* 0x90 */ "","&lsquo;","&rsquo;","&ldquo;","&rdquo;",
                                                 "&middot;","&ndash","&mdash",
  /* 0x98 */ "&tilde;","&trade;","&scaron;","&rang;","&oelig;","","","&Yuml;",
  /* 0xa0 */ "","&iexcl;","&cent;","&pound;",
                                         "&curren;","&yen;","&brvbar;","&sect",
  /* 0xa8 */ "&quot;","&copy;","a","&laquo;","&not;","-","&reg;","&macron;",
  /* 0xb0 */ "&deg;","&plusmn;","&sup2;","&sup3;",
                                    "&rsquot;","&micron;","&para;","&middot;",
  /* 0xb8 */ "&comma;","&sup1","&deg;","&raquo;",
                                  "&frac14;","&frac12;","&frac34;","&iquest;",
  /* 0xc0 */ "&Agrave;","&Aacute;","&Acirc;","&Atilde;",
                                      "&Auml;","&Aring;","&AElig;","&Ccedil;",
  /* 0xc8 */ "&Egrave;","&Eacute;","&Ecirc;","&Euml;",
                                     "&Igrave;","&Iacute;","&Icirc;","&Iuml;",
  /* 0xd0 */ "&ETH;","&Ntilde;","&Ograve;","&Oacute;",
                                      "&Ocirc;","&Otilde;","&Ouml;","&times;",
  /* 0xd8 */ "&Oslash;","&Ugrave;","&Uacute;","&Ucirc;",
                                      "&Uuml;","&Yacute;","&THORN;","&szlig;",
  /* 0xe0 */ "&agrave;","&aacute;","&acirc;","&atilde;",
                                      "&auml;","&aring;","&aelig;","&ccedil;",
  /* 0xe8 */ "&egrave;","&eacute;","&ecirc;","&euml;",
                                     "&igrave;","&iacute;","&icirc;","&iuml;",
  /* 0xf0 */ "&eth;","&ntilde;","&ograve;","&oacute;",
                                     "&ocirc;","&otilde;","&ouml;","&divide;",
  /* 0xf8 */ "&oslash;","&ugrave;","&uacute;","&ucirc;",
                                       "&uuml;","&yacute;","&thorn;","&yuml;"
};

static psiconv_character_layout gen_base_char(const psiconv_font font,
                                              const psiconv_color color,
                                              const psiconv_color back_color);
static void diff_char(FILE *of, const psiconv_character_layout old, 
                      const psiconv_character_layout new, int *flags);
static void gen_para(FILE *of, const psiconv_paragraph para, 
                     const psiconv_character_layout base_char);

static void psiconv_gen_html_word(FILE *of,psiconv_word_f wf);
static void psiconv_gen_html_texted(FILE *of,psiconv_texted_f tf);

/* This is not necessarily the same as returned by basic_character_layout_status
   This one is specific for the base point of HTML */
psiconv_character_layout gen_base_char(const psiconv_font font,
                                       const psiconv_color color,
                                       const psiconv_color back_color)
{
  struct psiconv_character_layout_s base_char_struct =
  {
    NULL,                  /* color               */
    NULL,                  /* back_color          */
    13.0,                  /* font_size           */
    psiconv_bool_false,    /* italic              */
    psiconv_bool_false,    /* bold                */
    psiconv_normalscript,  /* super_sub           */
    psiconv_bool_false,    /* underline           */
    psiconv_bool_false,    /* strikethrough       */
    NULL,                  /* font                */
  };
  base_char_struct.color = color;
  base_char_struct.back_color = back_color;
  base_char_struct.font = font;
  return psiconv_clone_character_layout(&base_char_struct);
}

/* flags & 1: 0 if no <FONT> was yet generated.
   flags & 2: 1 if at end-of-paragraph
*/
void diff_char(FILE *of, const psiconv_character_layout old, 
               const psiconv_character_layout new, 
               int *flags)
{
  int font_set = 0;

  if ((old->font_size != new->font_size) ||
      (old->color->red != new->color->red) ||
      (old->color->green != new->color->green) ||
      (old->color->blue != new->color->blue) ||
      (strcmp(old->font->name,new->font->name)) ||
      (old->font->screenfont != new->font->screenfont) ||
      ((*flags & 0x03) == 3)) {
    if (old->italic)
      fputs("</I>",of);
    if (old->bold)
      fputs("</B>",of);
    if (old->underline)
      fputs("</U>",of);
    if (old->strikethrough)
      fputs("</STRIKE>",of);
    if (old->super_sub == psiconv_superscript)
      fputs("</SUP>",of);
    if (old->super_sub == psiconv_subscript)
      fputs("</SUB>",of);
    if ((*flags & 1) == 1) 
      fputs("</FONT>",of);
    if ((*flags & 2) == 0) {
      *flags |= 1;
      fputs("<FONT SIZE=",of);
      if (new->font_size <= 8.0)
        fputs("1",of);
      else if (new->font_size <= 10.0)
        fputs("2",of);
      else if (new->font_size <= 12.0)
        fputs("3",of);
      else if (new->font_size <= 14.0)
        fputs("4",of);
      else if (new->font_size <= 18.0)
        fputs("5",of);
      else if (new->font_size <= 24.0)
        fputs("6",of);
      else 
        fputs("7",of);
      fprintf(of," COLOR=#%02x%02x%02x",new->color->red,new->color->green,
              new->color->blue);
      if (new->font->screenfont == psiconv_font_sansserif)
        fprintf(of," FACE=\"%s, Sans-Serif\">",new->font->name);
      else if (new->font->screenfont == psiconv_font_nonprop)
        fprintf(of," FACE=\"%s, Monospace\">",new->font-> name);
      else if (new->font->screenfont == psiconv_font_serif)
        fprintf(of," FACE=\"%s, Serif\">",new->font-> name);
      else
        fprintf(of," FACE=\"%s, Serif\">",new->font-> name);
    }
    if (new->italic)
      fputs("<I>",of);
    if (new->bold)
      fputs("<B>",of);
    if (new->underline)
      fputs("<U>",of);
    if (new->strikethrough)
      fputs("<STRIKE>",of);
    if (new->super_sub == psiconv_superscript)
      fputs("<SUP>",of);
    if (new->super_sub == psiconv_subscript)
      fputs("<SUB>",of);
  } else {
    if (font_set || (old->italic != new->italic)) {
      if (old->italic)
        fputs("</I>",of);
      else
        fputs("<I>",of);
    }
    if (old->bold != new->bold) {
      if (old->bold)
        fputs("</B>",of);
      else
        fputs("<B>",of);
    }
    if (old->underline != new->underline) {
      if (old->underline)
        fputs("</U>",of);
      else
        fputs("<U>",of);
    }
    if (old->strikethrough != new->strikethrough) {
      if (old->strikethrough)
        fputs("</STRIKE>",of);
      else
        fputs("<STRIKE>",of);
    }
    if (old->super_sub != new->super_sub) {
      if (old->super_sub == psiconv_superscript) 
        fputs("</SUP>",of);
      else if (old->super_sub == psiconv_subscript)
        fputs("</SUB>",of);
      if (new->super_sub == psiconv_superscript) 
        fputs("<SUP>",of);
      else if (new->super_sub == psiconv_subscript)
        fputs("<SUB>",of);
    }
  }
}

void gen_para(FILE *of, const psiconv_paragraph para, 
              const psiconv_character_layout base_char) 
{
  int i,j,loc;
  psiconv_character_layout cur_char;
  psiconv_in_line_layout inl;
  int flags = 0;


  fputs("<P",of);
  if (para->base_paragraph->justify_hor == psiconv_justify_left)
    fputs(" ALIGN=left",of);
  else if (para->base_paragraph->justify_hor == psiconv_justify_right)
    fputs(" ALIGN=right",of);
  else if (para->base_paragraph->justify_hor == psiconv_justify_centre)
    fputs(" ALIGN=center",of);
  else if (para->base_paragraph->justify_hor == psiconv_justify_full)
    fputs(" ALIGN=left",of);
  fputs(">",of);
  if (para->base_paragraph->bullet->on)
    fputs("<UL><LI>",of);

  cur_char = base_char;

  if (psiconv_list_length(para->in_lines) == 0) {
    diff_char(of,cur_char,para->base_character,&flags);
    cur_char = para->base_character;
  }
  loc = 0;

  for (i = 0; i < psiconv_list_length(para->in_lines); i++) {
    inl = psiconv_list_get(para->in_lines,i);
    diff_char(of,cur_char,inl->layout,&flags);
    cur_char = inl->layout;
    for (j = loc; j < inl->length + loc; j ++) {
      fputs(char_table[(unsigned char) (para->text[j])],of);
    }
    loc = j;
  }

  if (loc <  strlen(para->text)) {
    diff_char(of,cur_char,para->base_character,&flags);
    cur_char = para->base_character;
    for (j = loc; j < strlen(para->text); j ++) {
      fputs(char_table[(unsigned char) (para->text[j])],of);
    }
  }

  if (strlen(para->text) == 0) 
    fputs("<BR>",of);

  flags |= 2;
  diff_char(of,cur_char,base_char,&flags);

  if (para->base_paragraph->bullet->on)
    fputs("</UL>",of);
      
  fputs("</P>\n",of);
}

int psiconv_gen_html(const char * filename,const psiconv_file file, 
                     const char *dest)
{
  FILE *of = fopen(filename,"w");
  if (! of)
    return -1;

  if (file->type == psiconv_word_file) {
    psiconv_gen_html_word(of,(psiconv_word_f) file->file);
  } else if (file->type == psiconv_texted_file) {
    psiconv_gen_html_texted(of,(psiconv_texted_f) file->file);
  } else {
    fclose(of);
    return -1;
  }
  return fclose(of);
}

void psiconv_gen_html_texted(FILE *of,psiconv_texted_f tf)
{
  psiconv_character_layout base_char;
  psiconv_paragraph para;
  int i;

  /* We have nothing better */
  base_char = psiconv_basic_character_layout();

  fputs("<!doctype html public \"-//W3C//DTD HTML 3.2 Final//EN\">", of);
  fputs("\n<HTML>\n<HEAD>\n  <META NAME=\"GENERATOR\"", of);
  fputs(" CONTENT=\"psiconv-" VERSION "\">\n", of);
  fputs("<BODY>\n",of);
  for (i = 0; i < psiconv_list_length(tf->texted_sec->paragraphs); i++) {
    para = psiconv_list_get(tf->texted_sec->paragraphs,i);
    gen_para(of,para,base_char);
  }
  fputs("</BODY>\n</HTML>\n",of);
  psiconv_free_character_layout(base_char);
}


void psiconv_gen_html_word(FILE *of,psiconv_word_f wf)
{
  int i;
  psiconv_paragraph para;
  psiconv_color white,black;
  psiconv_character_layout base_char;

  white = malloc(sizeof(*white));
  black = malloc(sizeof(*black));
  white->red = 0x00;
  white->green = 0x00;
  white->blue = 0x00;
  black->red = 0xff;
  black->green = 0xff;
  black->blue = 0xff;

  /* To keep from generating a font desc for each line */
  base_char = gen_base_char(wf->styles_sec->normal->character->font,
                            black,white);

  psiconv_free_color(black);
  psiconv_free_color(white);

  fputs("<!doctype html public \"-//W3C//DTD HTML 3.2 Final//EN\">", of);
  fputs("\n<HTML>\n<HEAD>\n  <META NAME=\"GENERATOR\"", of);
  fputs(" CONTENT=\"psiconv-" VERSION "\">\n", of);
  fputs("<BODY>\n",of);

  for (i = 0; i < psiconv_list_length(wf->paragraphs); i++) {
    para = psiconv_list_get(wf->paragraphs,i);
    gen_para(of,para,base_char);
  }
  fputs("</BODY>\n</HTML>\n",of);
  psiconv_free_character_layout(base_char);
}

static struct psiconv_fileformat_s ff = 
  {
    "HTML3",
    "HTML 3.2, not verified so probably not completely compliant",
    &psiconv_gen_html
  };

void init_html(void)
{
  psiconv_list_add(fileformat_list,&ff);
}
