/*
 * gen_html4.c - Part of psiconv, a PSION 5 data formats converter
 * Copyright (c) 1999  Andrew Johnson <anjohnson@iee.org>
 * Portions Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "psiconv/data.h"
#include "psiconv/list.h"
#include "gen.h"
#include "psiconv.h"

#ifndef TRUE
#define TRUE (0==0)
#endif

#ifdef DMALLOC
#include "dmalloc.h"
#endif

/*
 * Various string tables for HTML4 settings
 */

/* How to display each character */
static const char *char_table[0x100] = {
  /* 0x00 */ "", "", "", "", "", "", "<BR>\n", "<BR>\n",
  /* 0x08 */ "\n<HR>\n", "\t", "", "", "", "", "", "",
  /* 0x10 */ " ", "", "", "", "", "", "", "",
  /* 0x18 */ "", "", "", "", "", "", "", "",
  /* 0x20 */ " ", "!", "&quot;", "#", "$", "%", "&amp;", "'",
  /* 0x28 */ "(", ")", "*", "+", ",", "-", ".", "/",
  /* 0x30 */ "0", "1", "2", "3", "4", "5", "6", "7",
  /* 0x38 */ "8", "9", ":", ";", "&lt;", "=", "&gt;", "?",
  /* 0x40 */ "@", "A", "B", "C", "D", "E", "F", "G",
  /* 0x48 */ "H", "I", "J", "K", "L", "M", "N", "O",
  /* 0x50 */ "P", "Q", "R", "S", "T", "U", "V", "W",
  /* 0x58 */ "X", "Y", "Z", "[", "\\", "]", "^", "_",
  /* 0x60 */ "`", "a", "b", "c", "d", "e", "f", "g",
  /* 0x68 */ "h", "i", "j", "k", "l", "m", "n", "o",
  /* 0x70 */ "p", "q", "r", "s", "t", "u", "v", "w",
  /* 0x78 */ "x", "y", "z", "{", "|", "}", "~", "",
  /* 0x80 */ "", "", "&#8218;", "&#402;",
             "&#8222;", "&#8230;", "&#8224;", "&#8225;",
  /* 0x88 */ "&#136;", "&#8240;", "&#352;", "<",
             "&#338;", "", "", "",
  /* 0x90 */ "", "&#8216;", "&#8217;", "&#8220;",
             "&#8221;","&#8226;", "&#8211;", "&#8212;",
  /* 0x98 */ "&#732;", "&#8482;", "&#353;", ">",
             "&#339;", "", "", "&#376;",
  /* 0xa0 */ "&nbsp;", "&iexcl;", "&cent;", "&pound;",
             "&curren;", "&yen;", "&brvbar;", "&sect;",
  /* 0xa8 */ "&uml;", "&copy;", "&ordf;", "&laquo;",
             "&not;", "&shy;", "&reg;", "&#175;",
  /* 0xb0 */ "&deg;", "&plusmn;", "&sup2;", "&sup3;",
             "&acute;", "&micro;", "&para;", "&middot;",
  /* 0xb8 */ "&cedil;", "&sup1;", "&ordm;", "&raquo;",
             "&frac14;", "&frac12;", "&frac34;", "&iquest;",
  /* 0xc0 */ "&Agrave;", "&Aacute;", "&Acirc;", "&Atilde;",
             "&Auml;", "&Aring;", "&AElig;", "&Ccedil;",
  /* 0xc8 */ "&Egrave;", "&Eacute;", "&Ecirc;", "&Euml;",
             "&Igrave;", "&Iacute;", "&Icirc;", "&Iuml;",
  /* 0xd0 */ "&ETH;", "&Ntilde;", "&Ograve;", "&Oacute;",
             "&Ocirc;", "&Otilde;", "&Ouml;", "&times;",
  /* 0xd8 */ "&Oslash;", "&Ugrave;", "&Uacute;", "&Ucirc;",
             "&Uuml;", "&Yacute;", "&THORN;", "&szlig;",
  /* 0xe0 */ "&agrave;", "&aacute;", "&acirc;", "&atilde;",
             "&auml;", "&aring;", "&aelig;", "&ccedil;",
  /* 0xe8 */ "&egrave;", "&eacute;", "&ecirc;", "&euml;",
             "&igrave;", "&iacute;", "&icirc;", "&iuml;",
  /* 0xf0 */ "&eth;", "&ntilde;", "&ograve;", "&oacute;",
             "&ocirc;", "&otilde;", "&ouml;", "&divide;",
  /* 0xf8 */ "&oslash;", "&ugrave;", "&uacute;", "&ucirc;",
             "&uuml;", "&yacute;", "&thorn;", "&yuml;"
};

/* The order of these must match the enum psiconv_border_kind (data.h) */
static const char *border_strings[] = {
    "none", "solid", "double", "dotted", "dashed", "dashed", "dotted"
};

/* The order of these must match the enum psiconv_justify (data.h) */
static const char *justify_strings[] = {
    "left", "center", "right", "justify"
};

/* The order of these must match the enum psiconv_super_sub (data.h) */
static const char *vertical_strings[] = {
    "baseline", "super", "sub"
};


/*
 * Support routines for particular formatting structures
 */

static void fput_name(FILE * of, const char *string) {
    int i;

    for (i = 0; i < strlen(string); i++)
	if (isalnum(string[i]))
	    putc(string[i], of);
}

static void fput_text(FILE * of, const char *text, int length, int *pwasspace) {
    int j;
    int space;
    
    for (j = 0; j < length; j++) {
	space = (text[j] == ' ');
	if (*pwasspace && space)
	    fputs("&nbsp;", of);
	else
	    fputs(char_table[(unsigned char) (text[j])], of);
	*pwasspace = space;
    }
}

static const char *screenfont_name(psiconv_u8 screenfont)
{
  if (screenfont == psiconv_font_sansserif)
    return "sans-serif";
  else if (screenfont == psiconv_font_nonprop)
    return "monospace";
  else if (screenfont == psiconv_font_serif)
    return "serif";
  else
    return "serif";
}

static const char *generic_font_family(const char *font_name) {
    static const char *sans_font_names[] = {
	"Arial", "Univers", "Sans Serif"
    };
    static const char *mono_font_names[] = {
	"Courier New", "Courier", "Mono", "Lineprinter", "Elite", "Swiss",
	"Courier (compressed)", "Courier (expanded)", "Letter Gothic"
    };
    int i;
    
    for (i=0; i < sizeof(sans_font_names) / sizeof(sans_font_names[0]); i++)
	if (!strcmp(font_name, sans_font_names[i]))
	    return "sans-serif";
    
    for (i=0; i < sizeof(mono_font_names) / sizeof(mono_font_names[0]); i++)
	if (!strcmp(font_name, mono_font_names[i]))
	    return "monospace";
    
    return "serif";
}

static int colors_different(
    const psiconv_color old,
    const psiconv_color new)
{
    return ((old->red != new->red) ||
	    (old->green != new->green) ||
	    (old->blue != new->blue));
}


/* Borders */
static int borders_different(
    const psiconv_border old, 
    const psiconv_border new)
{
    return (!old || 
    	    (old->kind != new->kind) ||
	    colors_different(old->color, new->color) ||
	    (old->thickness != new->thickness));
}

static void fput_border(FILE * of, 
    const char *which,
    const psiconv_border old, 
    const psiconv_border new)
{
    /* If the border changed at all we'll set all parameters, as they're
     * fairly closely linked and it's much easier to do it this way */
    if (borders_different(old, new)) {
	fprintf(of, "border-%s-style:%s;", 
		which, border_strings[new->kind]);
	if ((new->kind != psiconv_border_none) &&
    	    (new->color->red || new->color->green || new->color->blue))
	    fprintf(of, "border-%s-color:#%02x%02x%02x;", which,
		    new->color->red, new->color->green, new->color->blue);
	if ((new->kind == psiconv_border_solid) ||
	    (new->kind == psiconv_border_double))
	    fprintf(of, "border-%s-width:%.0fpx;", 
		    which, (new->thickness / 10.0) - 1);
	else if (new->kind != psiconv_border_none)
	    fprintf(of, "border-%s-width:1px;", which);
    }
}

/* Paragraph formats */
static int paragraph_layouts_different(
    const psiconv_paragraph_layout old,
    const psiconv_paragraph_layout new)
{
    return ((old->space_above != new->space_above) ||
    	    (old->indent_right != new->indent_right) ||
    	    (old->space_below != new->space_below) ||
    	    (old->indent_left != new->indent_left) ||
    	    (old->indent_first != new->indent_first) ||
    	    (old->justify_hor != new->justify_hor) ||
    	    (old->linespacing_exact != new->linespacing_exact) ||
	    borders_different(old->left_border, new->left_border) ||
	    borders_different(old->right_border, new->right_border) ||
	    borders_different(old->top_border, new->top_border) ||
	    borders_different(old->bottom_border, new->bottom_border));
}

static void fput_paragraph_layout(FILE * of, 
    const psiconv_paragraph_layout old,
    const psiconv_paragraph_layout new)
{
    if (!old || (old->space_above != new->space_above))
	fprintf(of, "margin-top:%.1fpt;", new->space_above);
    if (!old || (old->indent_right != new->indent_right))
	fprintf(of, "margin-right:%.2fcm;", new->indent_right);
    if (!old || (old->space_below != new->space_below))
	fprintf(of, "margin-bottom:%.1fpt;", new->space_below);
    if (!old || 
	(old->bullet->on != new->bullet->on) ||
	(old->bullet->indent != new->bullet->indent) ||
	(old->indent_left != new->indent_left) ||
	(old->indent_first != new->indent_first)) {
	psiconv_length_t indent = new->indent_first;
	psiconv_length_t margin = new->indent_left;
	/* Bullets with indent set cause strange margin behaviour: */
	if (new->bullet->on && new->bullet->indent) {
	    if (indent > 0.0) {
		margin += indent;
		indent = new->indent_left - indent;
	    }
	}
	fprintf(of, "margin-left:%.2fcm;", margin);
	fprintf(of, "text-indent:%.2fcm;", indent);
    }
    if (!old || (old->justify_hor != new->justify_hor))
	fprintf(of, "text-align:%s;", justify_strings[new->justify_hor]);
    if (!old || (old->linespacing_exact != new->linespacing_exact)) {
	if (!old || old->linespacing_exact)
	    fputs("line-height:normal;", of);
	else
	    fprintf(of, "line-height:%.1fpt;", new->linespacing);
    }
    fput_border(of, "left", old ? old->left_border : NULL, 
    		new->left_border);
    fput_border(of, "right", old ? old->right_border : NULL, 
    		new->right_border);
    fput_border(of, "top", old ? old->top_border : NULL, 
    		new->top_border);
    fput_border(of, "bottom", old ? old->bottom_border : NULL, 
    		new->bottom_border);
}

/* Character formats */
static int character_layouts_different(
    const psiconv_character_layout old,
    const psiconv_character_layout new)
{
    return ((old->font->screenfont != new->font->screenfont) ||
            (strcmp(old->font->name,new->font->name)) ||
	    (old->font_size != new->font_size) ||
	    (old->color->red != new->color->red) ||
	    (old->color->green != new->color->green) ||
	    (old->color->blue != new->color->blue) ||
	    (old->back_color->red != new->back_color->red) ||
	    (old->back_color->green != new->back_color->green) ||
	    (old->back_color->blue != new->back_color->blue) ||
	    (old->italic != new->italic) ||
	    (old->bold != new->bold) ||
	    (old->super_sub != new->super_sub) ||
	    (old->underline != new->underline) ||
	    (old->strikethrough != new->strikethrough));
}

static void fput_character_layout(FILE * of,
    const psiconv_character_layout old,
    const psiconv_character_layout new,
    psiconv_word_f wf)
{
    if (!old || (old->font->screenfont != new->font->screenfont) ||
        strcmp(old->font->name,new->font->name)) {
	   fprintf(of, "font-family:%s,%s,%s;", new->font->name,
	    	   generic_font_family(new->font->name),
                   screenfont_name(new->font->screenfont));
    }
    if (!old || (old->font_size != new->font_size))
    	fprintf(of, "font-size:%.1f%%;", 100.0 * new->font_size / 10.0);
    if (!old || (old->bold != new->bold))
    	fprintf(of, "font-weight:%s;", new->bold ? "bold" : "normal");
    if (!old || (old->italic != new->italic))
    	fprintf(of, "font-style:%s;", new->italic ? "italic" : "normal");
    if (!old || (colors_different(old->color, new->color)))
	fprintf(of, "color:#%02x%02x%02x;", 
	        new->color->red, new->color->green, new->color->blue);
    if (!old || (colors_different(old->back_color, new->back_color)))
	fprintf(of, "background-color:#%02x%02x%02x;",
		new->back_color->red, new->back_color->green, new->back_color->blue);
    if (!old || (old->super_sub != new->super_sub))
    	fprintf(of, "vertical-align:%s;", vertical_strings[new->super_sub]);
    if (!old || (old->underline != new->underline) ||
        (old->strikethrough != new->strikethrough)) {
	fputs("text-decoration:", of);
	if (new->strikethrough)
	    fputs("line-through ", of);
	if (new->underline)
	    fputs("underline", of);
	else if (!new->strikethrough)
	    fputs("none", of);
	fputs(";", of);
    }
}


/* Formatted paragraphs */
static void fput_para(FILE * of, 
    const psiconv_paragraph para, 
    psiconv_word_f wf)
{
    psiconv_paragraph_layout base_para = para->base_paragraph;
    psiconv_character_layout base_char = para->base_character;
    psiconv_character_layout cur_char;
    psiconv_character_layout *layout_stack;
    psiconv_in_line_layout inl;
    psiconv_word_style sty;
    int spans = 0;
    int wasspace = 0;
    int i, j, loc;
    
    layout_stack = calloc(psiconv_list_length(para->in_lines),
    			  sizeof(*layout_stack));
    fputs("<P", of);
    sty = psiconv_get_style(wf->styles_sec, para->base_style);
    cur_char = sty->character;
    
    if (sty->name) {
	fputs(" CLASS=\"", of);
	fput_name(of, sty->name);
	putc('"', of);
    }
    
    if (paragraph_layouts_different(sty->paragraph, base_para) ||
        character_layouts_different(cur_char, base_char)) {
	fputs(" STYLE=\"", of);
	fput_paragraph_layout(of, sty->paragraph, base_para);
	fput_character_layout(of, cur_char, base_char, wf);
	cur_char = base_char;
	putc('"', of);
    }
    putc('>', of);
    
    if (base_para->bullet->on) {
	fputs("<SPAN STYLE=\"", of);
	if (base_para->bullet->font != base_char->font) {
	    fprintf(of, "font-family:%s,%s,%s;", 
	    	    base_para->bullet->font->name, 
                    generic_font_family(base_para->bullet->font->name),
                    screenfont_name(base_para->bullet->font->screenfont));
	}
	if (base_para->bullet->font_size != base_char->font_size)
	    fprintf(of, "font-size:%.1f%%;", 
		    100.0 * base_para->bullet->font_size / 10.0);
	if ((base_para->bullet->color->red != base_char->color->red) ||
	    (base_para->bullet->color->green != base_char->color->green) ||
	    (base_para->bullet->color->blue != base_char->color->blue))
	    fprintf(of, "color:#%02x%02x%02x;", 
	            base_para->bullet->color->red, 
		    base_para->bullet->color->green, 
		    base_para->bullet->color->blue);
	fprintf(of, "\">%s </SPAN>", 
		char_table[base_para->bullet->character]);
	wasspace = TRUE;
    }
    
    loc = 0;
    for (i = 0; i < psiconv_list_length(para->in_lines); i++) {
	inl = psiconv_list_get(para->in_lines, i);
	if (character_layouts_different(cur_char, inl->layout)) {
	    for (j = 0; j < spans; j++)
		if (!character_layouts_different(inl->layout, layout_stack[j]))
		    break;
	    if (j < spans) {
		while (spans > j) {
		    fputs("</SPAN>",of);
		    spans--;
		}
	    } else {
		fputs("<SPAN STYLE=\"", of);
		fput_character_layout(of, cur_char, inl->layout, wf);
		fputs("\">",of);
		layout_stack[spans++] = cur_char;
	    }
	    cur_char = inl->layout;
	}
	fput_text(of, &para->text[loc], inl->length, &wasspace);
	loc += inl->length;
    }
    
    while (spans > 0) {
        fputs("</SPAN>",of);
	spans--;
    }

    if (loc < strlen(para->text)) {
	fput_text(of, &para->text[loc], strlen(para->text) - loc, &wasspace);
    }
    
    if (strlen(para->text) == 0)
	fputs("&nbsp;", of);

    fputs("</P>\n", of);
    free(layout_stack);
}

static void fput_style(FILE * of, 
    const psiconv_word_style normal, 
    const psiconv_word_style sty, 
    psiconv_word_f wf)
{
    if (sty->name == NULL)
	fputs("  P {", of);
    else {
	fputs("  P.", of);
	fput_name(of, sty->name);
	fputs(" {", of);
    }
    fput_paragraph_layout(of, normal ? normal->paragraph : NULL, sty->paragraph);
    fput_character_layout(of, normal ? normal->character : NULL, sty->character, wf);
    fputs("}\n", of);
}

static int psiconv_gen_html4(const char *filename, const psiconv_file file, 
                             const char *dest)
{
    FILE * of;
    int i;
    psiconv_paragraph para;
    psiconv_word_style normal, sty;
    psiconv_word_f wf;

    if (file->type == psiconv_word_file)
      wf = file->file;
    else {
      /* Fall back on the normal HTML generator */
      return psiconv_gen_html(filename,file,dest);
    }

    if (! (of = fopen(filename,"w"))) 
        return -1;

    fputs("<!doctype html public \"-//W3C//DTD HTML 4.0 STRICT//EN\">", of);
    fputs("\n<HTML>\n<HEAD>\n  <META NAME=\"GENERATOR\"", of);
    fputs(" CONTENT=\"psiconv-" VERSION "\">\n", of);
    
    if (wf->page_sec->header->text) {
        if (psiconv_list_length(wf->page_sec->header->text->paragraphs) > 0) {
    	    fputs("  <TITLE>", of);
	    para = psiconv_list_get(wf->page_sec->header->text->paragraphs, 0);
	    i = 0;
	    fput_text(of, para->text, strlen(para->text), &i);
	    fputs("</TITLE>\n", of);
        }
    }
    
    normal = wf->styles_sec->normal;
    fputs("  <STYLE TYPE=\"text/css\"><!--\n", of);
    fprintf(of,"  BODY { background-color: #%02x%02x%02x }\n",
    	  normal->paragraph->back_color->red, 
	  normal->paragraph->back_color->green,
	  normal->paragraph->back_color->blue);
    fput_style(of, NULL, normal, wf);
    for (i = 0; i < psiconv_list_length(wf->styles_sec->styles); i++) {
	sty = psiconv_list_get(wf->styles_sec->styles, i);
	fput_style(of, normal, sty, wf);
    }
    fputs("  --></STYLE>\n</HEAD>\n<BODY>\n", of);
    if (wf->page_sec->header->text) {
        for (i = 0; i < psiconv_list_length(wf->paragraphs); i++) {
	    para = psiconv_list_get(wf->paragraphs, i);
	    fput_para(of, para, wf);
        }
    }
    fputs("</BODY>\n</HTML>\n", of);

    return fclose(of);
}

static struct psiconv_fileformat_s ff =
  {
    "HTML4",
    "HTML 4.0 with cascading style sheets",
    psiconv_gen_html4
  };

void init_html4(void)
{
  psiconv_list_add(fileformat_list,&ff);
}

