/*
    gen_latex.h - Part of psiconv, a PSION 5 file formats converter
    Copyright (c) 2001  Jim Ottaway <j.ottaway@lse.ac.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* gen_latex.c

   Sun Feb 11 12:16:29 2001

   Jim Ottaway <j.ottaway@lse.ac.uk>

   Conversion to LaTeX:

   * Converts to the article class

   * If the file is a Word file and has outline levels (i.e. is not a
   Psion 5 file), outline levels 1 to 5 are converted to sectioning
   commands, otherwise headings are just formatted

   * If there is a style with the name 'quotation', the paragraph is
   converted to a quotation environment (other styles/environments
   could be added)

   * Also does formatting commands (italics -> \emph, bold -> \textbf,
   underline -> \underline), and character translation

*/

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "psiconv/data.h"
#include "psiconv/list.h"
#include "gen.h"
#include "psiconv.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif

/* This is incomplete at the moment.
   
   Most of the translation/faking of chars is borrowed from the
   HTML::Latex.pm perl module. */
static const char *char_table[0x100] =
{
  /* 0x00 */ ""     ,""     ,""     ,""     ,""     ,""     ,"\n\n","\\\\"
,
  /* 0x08 */ "\n\n"  ," "    ,""     ,""     ,""     ,""     ,""     ,""
,
  /* 0x10 */ " "    ,""     ,""     ,""     ,""     ,""     ,""     ,""
,
  /* 0x18 */ ""     ,""     ,""     ,""     ,""     ,""     ,""     ,""
,
  /* 0x20 */ " "    ,"!"    ,"\""   ,"\\#"   ,"\\$"    ,"\\%"    ,"\\&","'"
,
  /* 0x28 */ "("    ,")"    ,"*"    ,"+"    ,","    ,"-"    ,"."    ,"/"
,
  /* 0x30 */ "0"    ,"1"    ,"2"    ,"3"    ,"4"    ,"5"    ,"6"    ,"7"
,
  /* 0x38 */ "8"    ,"9"    ,":"    ,";"    ,"$<$"  ,"="   ,"$>$" ,"?"     ,
  /* 0x40 */ "@"    ,"A"    ,"B"    ,"C"    ,"D"    ,"E"    ,"F"    ,"G"
,
  /* 0x48 */ "H"    ,"I"    ,"J"    ,"K"    ,"L"    ,"M"    ,"N"    ,"O"
,
  /* 0x50 */ "P"    ,"Q"    ,"R"    ,"S"    ,"T"    ,"U"    ,"V"    ,"W"
,
  /* 0x58 */ "X"     ,"Y"   ,"Z"    ,"["    ,"$\\backslash$" ,"]" ,"\\^{}"
,"\\_" ,
  /* 0x60 */ "`"    ,"a"    ,"b"    ,"c"    ,"d"    ,"e"    ,"f"    ,"g"
,
  /* 0x68 */ "h"    ,"i"    ,"j"    ,"k"    ,"l"    ,"m"    ,"n"    ,"o"
,
  /* 0x70 */ "p"    ,"q"    ,"r"    ,"s"    ,"t"    ,"u"    ,"v"    ,"w"
,
  /* 0x78 */ "x"    ,"y"    ,"z"    ,"\\{"    ,"$|$"    ,"\\}"    ,"~"
,""      ,
  /* 0x80 */ ""     ,"","","$f$","","{\\ldots}", "$\\dagger$","$\\ddagger$",
  /* 0x88 */ "\\^{}","","","","\\OE","" ,""    ,""      ,
  /* 0x90 */ "","`","'","``","''", "$\\bullet$","--","---",
  /* 0x98 */ "~","{\\textsc{tm}}","","","\\oe","","","\\\"Y",
  /* 0xa0 */ "","!`","c","{\\pounds}",
"","{Y\\hspace*{-1.4ex}--}","$|$","{\\S}",
  /* 0xa8 */
"\"","{\\copyright}","$^{\\underline{a}}$","","$\\neg$","$-$","","$^-$",
  /* 0xb0 */ "$^{\\circ}$","$\\pm$","$^2$","$^3$",
"$^\\prime$","$\\mu$","{\\P}","$\\cdot$",
  /* 0xb8 */ ",","$^1$","$^{\\underline{\\circ}}$","",
"$\\frac{1}{4}$","$\\frac{1}{2}$","$\\frac{3}{4}$","?`",
  /* 0xc0 */ "\\`A","\\'A","\\^A","\\~A", "\\\"A","\\AA","\\AE","\\c{C}",
  /* 0xc8 */ "\\`E","\\'E","\\^E","\\\"E", "\\`I","\\'I","\\^I","\\\"I",
  /* 0xd0 */ "{D\\hspace*{-1.7ex}-\\hspace{.9ex}}","\\~N","\\`O","\\'O",
"\\^O","\\~O","\\\"O","$\\times$",
  /* 0xd8 */ "{\\O}","\\`U","\\'U","\\^U", "\\\"U","\\'Y","","",
  /* 0xe0 */ "\\`a","\\'a","\\^a","\\~a",
"\\\"a","\\r{a}","{\\ae}","\\c{c}",
  /* 0xe8 */ "\\`e","\\'e","\\^e","\\\"e",
"\\`{\\i}","\\'{\\i}","\\^{\\i}","\\\"{\\i}",
  /* 0xf0 */ "\\v{o}","\\~n","\\`o","\\'o", "\\^o","\\~o","\\\"o","$\\div$",
  /* 0xf8 */ "{\\o}","\\`u","\\'u","\\^u", "\\\"u","\\'y","","\\\"y"
};

static psiconv_character_layout gen_base_char(const psiconv_font font,
                                              const psiconv_color color,
                                              const psiconv_color
back_color);
static void diff_char(FILE *of, const psiconv_character_layout old,
                      const psiconv_character_layout new, int *flags);
static void gen_para(FILE *of, const psiconv_paragraph para,
                     const psiconv_character_layout base_char,
		     psiconv_word_f wf);

static void psiconv_gen_latex_word(FILE *of,psiconv_word_f wf);
static void psiconv_gen_latex_texted(FILE *of,psiconv_texted_f tf);

psiconv_character_layout gen_base_char(const psiconv_font font,
                                       const psiconv_color color,
                                       const psiconv_color back_color)
{
  struct psiconv_character_layout_s base_char_struct =
  {
    NULL,                  /* color               */
    NULL,                  /* back_color          */
    13.0,                  /* font_size           */
    psiconv_bool_false,    /* italic              */
    psiconv_bool_false,    /* bold                */
    psiconv_normalscript,  /* super_sub           */
    psiconv_bool_false,    /* underline           */
    psiconv_bool_false,    /* strikethrough       */
    NULL,                  /* font                */
  };
  base_char_struct.color = color;
  base_char_struct.back_color = back_color;
  base_char_struct.font = font;
  return psiconv_clone_character_layout(&base_char_struct);
}

/* flags & 1: 1 if in a section
   flags & 2: 1 if at end-of-paragraph
*/
void diff_char(FILE *of, const psiconv_character_layout old,
               const psiconv_character_layout new,
               int *flags)
{
  if ((*flags & 3) == 3) {	/* end of section command argument */
    putc('}',of);
    return;
  }
  if (old->italic != new->italic) {
    if (old->italic)
      putc('}',of);
    else
      fputs("\\emph{",of);
  }
  if (old->bold != new->bold) {
    if (old->bold)
      putc('}',of);
    else
      fputs("\\textbf{",of);
  }
  if (old->underline != new->underline) {
    if (old->underline)
      putc('}',of);
    else
      fputs("\\underline{",of);
  }
}


const static char *sections[] = {
  "section",
  "subsection",
  "subsubsection",
  "paragraph",
  "subparagraph"
};

struct environment {
   char		*style_name;
   char		*environment_name;
};

const static struct environment environments[] = {
  {"quotation", 	"quotation"},
    {"",       		NULL}
};

char *find_env(char *style) {
  int n = 0;
   while (environments[n].environment_name != NULL) {
     if (strcmp(style, environments[n].style_name) == 0) {
       return environments[n].environment_name;
     }
     n++;
   }
   return NULL;
}

psiconv_bool_t bullet_switch_on = psiconv_bool_false;

void gen_para(FILE *of, const psiconv_paragraph para,
              const psiconv_character_layout base_char,
	      psiconv_word_f wf)
{
  int i,j,loc;
  psiconv_character_layout cur_char;
  psiconv_in_line_layout inl;
  int flags = 0;
  psiconv_word_style sty;
  char *env = NULL;

  if (para->base_paragraph->bullet->on) {
    if (! bullet_switch_on) {
      fputs("\\begin{itemize}\n\n", of);
      bullet_switch_on = psiconv_bool_true;
    }
    fputs("\\item ",of);
  } else {
    if (bullet_switch_on) {
      fputs("\\end{itemize}\n\n", of);
      bullet_switch_on = psiconv_bool_false;
    }
  }

  cur_char = base_char;

  if (wf) {
    sty = psiconv_get_style(wf->styles_sec, para->base_style);
    if (sty->name && (env = find_env(sty->name))) {
      fputs("\\begin{",of);
      fputs(env,of);
      fputs("}\n",of);
    } else {
      if (sty->outline_level &&
	  (sty->outline_level > 0) && (sty->outline_level < 6)) {
	putc('\\', of);
	fputs(sections[(sty->outline_level - 1)], of);
	putc('{', of);
	cur_char = para->base_character; /* ignore initial formatting */
	flags |= 1;
      }
    }
  }

  if (psiconv_list_length(para->in_lines) == 0) {
    diff_char(of,cur_char,para->base_character,&flags);
    cur_char = para->base_character;
  }
  loc = 0;

  for (i = 0; i < psiconv_list_length(para->in_lines); i++) {
    inl = psiconv_list_get(para->in_lines,i);
    diff_char(of,cur_char,inl->layout,&flags);
    cur_char = inl->layout;
    for (j = loc; j < inl->length + loc; j ++) {
      fputs(char_table[(unsigned char) (para->text[j])],of);
    }
    loc = j;
  }

  if (loc <  strlen(para->text)) {
    diff_char(of,cur_char,para->base_character,&flags);
    cur_char = para->base_character;
    for (j = loc; j < strlen(para->text); j ++) {
      fputs(char_table[(unsigned char) (para->text[j])],of);
    }
  }

  flags |= 2;
  diff_char(of,cur_char,base_char,&flags);

  if (env) {
    fputs("\n\\end{",of);
    fputs(env,of);
    putc('}',of);
  }

  fputs("\n\n", of);
}

int psiconv_gen_latex(const char * filename,const psiconv_file file,
                     const char *dest)
{
  FILE *of = fopen(filename,"w");
  if (! of)
    return -1;

  if (file->type == psiconv_word_file) {
    psiconv_gen_latex_word(of,(psiconv_word_f) file->file);
  } else if (file->type == psiconv_texted_file) {
    psiconv_gen_latex_texted(of,(psiconv_texted_f) file->file);
  } else {
    fclose(of);
    return -1;
  }
  return fclose(of);
}

/* This isn't tested !!! */
void psiconv_gen_latex_texted(FILE *of,psiconv_texted_f tf)
{
  psiconv_character_layout base_char;
  psiconv_paragraph para;
  int i;

  /* We have nothing better */
  base_char = psiconv_basic_character_layout();

  fputs("\\documentclass{article}\n\n\\begin{document}\n\n", of);
  for (i = 0; i < psiconv_list_length(tf->texted_sec->paragraphs); i++) {
    para = psiconv_list_get(tf->texted_sec->paragraphs,i);
    gen_para(of,para,base_char, NULL);
  }
  fputs("\\end{document}",of);
  psiconv_free_character_layout(base_char);
}

void psiconv_gen_latex_word(FILE *of,psiconv_word_f wf)
{
  int i;
  psiconv_paragraph para;
  psiconv_color white,black;
  psiconv_character_layout base_char;

  white = malloc(sizeof(*white));
  black = malloc(sizeof(*black));
  white->red = 0x00;
  white->green = 0x00;
  white->blue = 0x00;
  black->red = 0xff;
  black->green = 0xff;
  black->blue = 0xff;

  /* To keep from generating a font desc for each line */
  base_char = gen_base_char(wf->styles_sec->normal->character->font,
                            black,white);

  psiconv_free_color(black);
  psiconv_free_color(white);

  fputs("\\documentclass{article}\n\n\\begin{document}\n\n", of);

  for (i = 0; i < psiconv_list_length(wf->paragraphs); i++) {
    para = psiconv_list_get(wf->paragraphs,i);
    gen_para(of,para,base_char,wf);
  }
  fputs("\\end{document}\n",of);
  for (i = 0; i <
psiconv_list_length(wf->page_sec->header->text->paragraphs); i++) {
    para = psiconv_list_get(wf->page_sec->header->text->paragraphs,i);
    gen_para(of,para,base_char,wf);
  }

  psiconv_free_character_layout(base_char);
}

static struct psiconv_fileformat_s ff =
  {
    "LaTeX",
    "LaTeX conversion to article class",
    &psiconv_gen_latex
  };

void init_latex(void)
{
  psiconv_list_add(fileformat_list,&ff);
}

