/*
 * gen_text.c - Part of psiconv, a PSION 5 file formats converter
 * Copyright (c) 1999  Andrew Johnson <anjohnson@iee.org>
 * Portions Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include <stdio.h>
#include <string.h>
#include "psiconv/data.h"
#include "psiconv/list.h"
#include "gen.h"
#include "psiconv.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif


/*
 * Various string tables for HTML4 settings
 */

/* Character conversion table */
static const char *char_table[0x100] = {
  /* 0x00 */ "",     "",     "",     "",     "",     "",     "\n",   "\n",
  /* 0x08 */ "\n",   "\t",   "",     "",     "",     "",     "",     "",     
  /* 0x10 */ " ",    "",     "",     "",     "",     "",     "",     "",     
  /* 0x18 */ "",     "",     "",     "",     "",     "",     "",     "",     
  /* 0x20 */ " ",    "!",    "\"",   "#",    "$",    "%",    "&",    "'",
  /* 0x28 */ "(",    ")",    "*",    "+",    ",",    "-",    ".",    "/",
  /* 0x30 */ "0",    "1",    "2",    "3",    "4",    "5",    "6",    "7",
  /* 0x38 */ "8",    "9",    ":",    ";",    "<",    "=",    ">",    "?",
  /* 0x40 */ "@",    "A",    "B",    "C",    "D",    "E",    "F",    "G",
  /* 0x48 */ "H",    "I",    "J",    "K",    "L",    "M",    "N",    "O",
  /* 0x50 */ "P",    "Q",    "R",    "S",    "T",    "U",    "V",    "W",
  /* 0x58 */ "X",    "Y",    "Z",    "[",    "\\",   "]",    "^",    "_",
  /* 0x60 */ "`",    "a",    "b",    "c",    "d",    "e",    "f",    "g",
  /* 0x68 */ "h",    "i",    "j",    "k",    "l",    "m",    "n",    "o",
  /* 0x70 */ "p",    "q",    "r",    "s",    "t",    "u",    "v",    "w",
  /* 0x78 */ "x",    "y",    "z",    "{",    "|",    "}",    "~",    "",     
  /* 0x80 */ "",     "",     ",",    "f",    ",,",   "...",  "+",    "#",
  /* 0x88 */ "^",    "\176/oo","S",  "<",    "OE",   "",     "",     "",     
  /* 0x90 */ "",     "`",    "'",    "``",   "''",   "*",    "-",    "--",
  /* 0x98 */ "~",    "(TM)", "s",    ">",    "oe",   "",     "",     "Y",
  /* 0xa0 */ "\xa0", "\xa1", "\xa2", "\xa3", "\xa4", "\xa5", "\xa6", "\xa7",
  /* 0xa8 */ "\xa8", "\xa9", "\xaa", "\xab", "\xac", "\xad", "\xae", "\xaf",
  /* 0xb0 */ "\xb0", "\xb1", "\xb2", "\xb3", "\xb4", "\xb5", "\xb6", "\xb7",
  /* 0xb8 */ "\xb8", "\xb9", "\xba", "\xbb", "\xbc", "\xbd", "\xbe", "\xbf",
  /* 0xc0 */ "\xc0", "\xc1", "\xc2", "\xc3", "\xc4", "\xc5", "\xc6", "\xc7",
  /* 0xc8 */ "\xc8", "\xc9", "\xca", "\xcb", "\xcc", "\xcd", "\xce", "\xcf",
  /* 0xd0 */ "\xd0", "\xd1", "\xd2", "\xd3", "\xd4", "\xd5", "\xd6", "\xd7",
  /* 0xd8 */ "\xd8", "\xd9", "\xda", "\xdb", "\xdc", "\xdd", "\xde", "\xdf",
  /* 0xe0 */ "\xe0", "\xe1", "\xe2", "\xe3", "\xe4", "\xe5", "\xe6", "\xe7",
  /* 0xe8 */ "\xe8", "\xe9", "\xea", "\xeb", "\xec", "\xed", "\xee", "\xef",
  /* 0xf0 */ "\xf0", "\xf1", "\xf2", "\xf3", "\xf4", "\xf5", "\xf6", "\xf7",
  /* 0xf8 */ "\xf8", "\xf9", "\xfa", "\xfb", "\xfc", "\xfd", "\xfe", "\xff",
};


/* Output a string, doing character conversions */
static void fput_text(FILE * of, const char *text, int length) {
    int j;
    
    for (j = 0; j < length; j++) {
	fputs(char_table[(unsigned char) (text[j])], of);
    }
}


/* Output a paragraph */
static void fput_para(FILE * of, 
    const psiconv_paragraph para,int extra_nl)
{
    if (para->base_paragraph->bullet->on) {
	fprintf(of, "%s ", char_table[para->base_paragraph->bullet->character]);
    }
    fput_text(of, para->text, strlen(para->text));
    fputs("\n", of);
    if (extra_nl)
      fputs("\n", of);
}

static void psiconv_gen_txt_texted(FILE * of, psiconv_texted_f tf)
{
    int i;
    psiconv_paragraph para;

    if (tf->page_sec->header->text) {
	if (tf->page_sec->header->on_first_page) {
	    for (i=0; i < psiconv_list_length(tf->page_sec->header->text->paragraphs); i++) {
		para = psiconv_list_get(tf->page_sec->header->text->paragraphs, i);
		fput_text(of, para->text, strlen(para->text));
		fputs("\n", of);
 	    }
	}
    }
   fputs("\n",of);
    
    for (i=0; i < psiconv_list_length(tf->texted_sec->paragraphs); i++) {
	para = psiconv_list_get(tf->texted_sec->paragraphs, i);
	fput_para(of, para,0);
    }
    
   fputs("\n",of);
    if (tf->page_sec->header->text) {
        for (i=0; i < psiconv_list_length(tf->page_sec->footer->text->paragraphs); i++) {
		para = psiconv_list_get(tf->page_sec->footer->text->paragraphs, i);
		fput_text(of, para->text, strlen(para->text));
		fputs("\n", of);
        }
    }
}

static void psiconv_gen_txt_word(FILE * of, psiconv_word_f wf)
{
    int i;
    psiconv_paragraph para;
    
    if (wf->page_sec->header->on_first_page) {
	for (i=0; i < psiconv_list_length(wf->page_sec->header->text->paragraphs); i++) {
	    para = psiconv_list_get(wf->page_sec->header->text->paragraphs, i);
	    fput_text(of, para->text, strlen(para->text));
	    fputs("\n\n", of);
	}
    }
    fputs("\n\n", of);
    
    for (i=0; i < psiconv_list_length(wf->paragraphs); i++) {
	para = psiconv_list_get(wf->paragraphs, i);
	fput_para(of, para,1);
    }
    
    fputs("\n\n", of);
    for (i=0; i < psiconv_list_length(wf->page_sec->footer->text->paragraphs); i++) {
	para = psiconv_list_get(wf->page_sec->footer->text->paragraphs, i);
	fput_text(of, para->text, strlen(para->text));
	fputs("\n\n", of);
    }
}

static int psiconv_gen_txt(const char *filename, const psiconv_file file, 
                            const char *dest)
{
  FILE *of = fopen(filename,"w");
  if (! of)
    return -1;

  if (file->type == psiconv_word_file) {
    psiconv_gen_txt_word(of,(psiconv_word_f) file->file);
  } else if (file->type == psiconv_texted_file) {
    psiconv_gen_txt_texted(of,(psiconv_texted_f) file->file);
  } else {
    fclose(of);
    return -1;
  }
  return fclose(of);
}

static struct psiconv_fileformat_s ff =
  {
    "ASCII",
    "Plain text without much layout",
    psiconv_gen_txt
  };

void init_txt(void)
{
  psiconv_list_add(fileformat_list,&ff);
}

