/*
    psiconv.c - Part of psiconv, a PSION 5 file formats converter
    Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Driver program */

#include "config.h"
#include "compat.h"
#include <getopt.h>
#include <stdio.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef IMAGEMAGICK
#include <magick/magick.h>
#endif

#include "psiconv/data.h"
#include "psiconv/parse.h"
#include "gen.h"
#include "psiconv.h"

static void print_help(void);
static void print_version(void);
static void strtoupper(char *str);

void print_help(void)
{
  psiconv_fileformat ff;
  int i,j;

  puts("Syntax: psiconv [OPTIONS..] [FILE]");
  puts("Convert the psion Word file FILE to a HTML file");
  puts("If FILE is not specified, use stdin");
  puts("  -d, --debug          Show debug information on stderr");
  puts("  -h, --help           Display this help and exit");
  puts("  -o, --outputfile     Output to file instead of stdout");
  puts("  -s, --silent         Do not even show warnings on stderr");
  puts("  -T, --type=FILETYPE  Output type");
  puts("  -V, --version        Display the program version and exit");
  puts("  -v, --verbose        Show progress indicators on stderr");
  puts("");
  puts("The following output types are known:");
  for (i = 0; i < psiconv_list_length(fileformat_list); i ++) {
    ff = psiconv_list_get(fileformat_list,i);
    printf("  %s",ff->name);
    for (j = strlen(ff->name); j < 15; j++)
      putchar(' ');
    puts(ff->description);
  }
}

void print_version(void)
{
  printf("Version %s\n",VERSION);
}

void strtoupper(char *str)
{
  int i;
  for (i = 0; i < strlen(str); i ++) 
    str[i] = toupper(str[i]);
}

int main(int argc, char *argv[])
{
  struct option long_options[] =
  {
    {"help",no_argument,NULL,'h'},
    {"version",no_argument,NULL,'V'},
    {"verbose",no_argument,NULL,'v'},
    {"debug",no_argument,NULL,'d'},
    {"silent",no_argument,NULL,'s'},
    {"outputfile",required_argument,NULL,'o'},
    {"type",required_argument,NULL,'T'},
    {0,0,0,0}
  };
  const char* short_options = "hVvsdo:eT:";
  int option_index;
  FILE * f;
  struct stat fbuf;

  const char *inputfilename = "";
  const char *outputfilename = "";
  char *type = strdup("HTML3");

  int c,i,res;
  psiconv_buffer buf;
  psiconv_file file;
  psiconv_fileformat ff;

  psiconv_verbosity = PSICONV_VERB_WARN;

  fileformat_list = psiconv_list_new(sizeof(struct psiconv_fileformat_s));
  init_txt();
  init_html();
  init_html4();
  init_latex();
  init_rtf();
  init_image();

  while(1) {
    c = getopt_long(argc,argv,short_options, long_options, &option_index);
    if (c == -1)
      break;
    switch(c) {
      case 'h': print_help(); exit(0);
      case 'V': print_version(); exit(0);
      case 'v': if (psiconv_verbosity < PSICONV_VERB_PROGRESS) 
                  psiconv_verbosity = PSICONV_VERB_PROGRESS; 
                break;
      case 'd': psiconv_verbosity = PSICONV_VERB_DEBUG; break;
      case 's': psiconv_verbosity = PSICONV_VERB_FATAL; break;
      case 'o': outputfilename = strdup(optarg); break;
      case 'T': type = strdup(optarg); break;
      case '?': case ':': fputs("Try `-h' for more information\n",stderr); 
                          exit(1);
      default: fprintf(stderr,"Internal error: getopt_long returned character "
                              "code 0%o ?? (contact the author)\n", c); 
               exit(1); break;
    }
  }
  if (optind < argc-1) {
    fputs("I can only convert one file!\n"
          "Try `-h' for more information\n",stderr);
    exit(1);
  } else if (optind == argc-1) 
    inputfilename = strdup(argv[optind]);


  /* Open inputfile for reading */
    
  if (strlen(inputfilename) != 0) {
    if(stat(inputfilename,&fbuf) < 0) {
      perror(inputfilename);
      exit(1);
    }
    f = fopen(inputfilename,"r");
    if (! f) {
      perror(inputfilename);
      exit(1);
    }
  } else 
    f = stdin;

  buf = psiconv_buffer_new();
  psiconv_buffer_fread_all(buf,f);

  if (strlen(inputfilename) != 0) 
    if (fclose(f)) {
      perror(inputfilename);
      exit(1);
    }

  if (psiconv_parse(buf,&file) || (file->type == psiconv_unknown_file))
  {
     fprintf(stderr,"Parse error\n");
     exit(1);
  }

  /* Set correct output file */
  if (strlen(outputfilename) == 0) 
    outputfilename = "/dev/stdout";

  strtoupper(type);
  for (i = 0; i < psiconv_list_length(fileformat_list); i ++) {
    ff = psiconv_list_get(fileformat_list,i);
    if (! strcmp(type,ff->name)) {
      res = ff->output(outputfilename,file,type);
      if (res) {
        fprintf(stderr,
                "Output format `%s' not permitted for this file type\n",type);
        exit(1);
      }
      break;
    }
  }

  if (i == psiconv_list_length(fileformat_list)) {
    fprintf(stderr,"Unknown output type: `%s'\n",type);
    exit(1);
  }

  psiconv_free_file(file);

  exit(0);
}
